local mod = ...



-- ************************************************************************************************
-- Component: Dragon ------------------------------------------------------------------------------
-- ************************************************************************************************

-- Sort a table by column "dist" -------------------------------------------------------------------------------------
function SortByDist(row1, row2)
    return row1.dist < row2.dist
end


-- Ease value at the extremes
local function easeValue(t)
    local ease = math.abs(1 - (math.cos(t * math.pi) / 2))
    ease = 1 - ease + 0.25
    -- if ease > -0.1 and ease < 0 then
    --     ease = -0.1
    -- elseif ease > 0 and ease < 0.1 then
    --     ease = 0.1
    -- end
    if ease < 0.1 then
        ease = 0.1
    elseif ease > 0.75 then
        ease = 0.75
    end

    -- -- local ease = 1
    return ease
end


-- Component
local COMP_DRAGON = {
	TypeName = "COMP_DRAGON",
	ParentType = "COMPONENT",
	Properties = {
		{ Name = "InitFlag", Type = "boolean", Default = false, Flags = { "SAVE_GAME" } },
		{ Name = "MovementSpeed", Type = "float", Default = 6.0, Flags = { "SAVE_GAME" } }, -- 6.0
		{ Name = "AnimationSpeed", Type = "float", Default = 7.0, Flags = { "SAVE_GAME" } }, -- 7.0
		{ Name = "WingAngle", Type = "float", Default = 35.0, Flags = { "SAVE_GAME" } },
		{ Name = "WholeFactor", Type = "float", Default = 0.10, Flags = { "SAVE_GAME" } },
		{ Name = "HeadFactor", Type = "float", Default = 0.20, Flags = { "SAVE_GAME" } },
		{ Name = "TailFactor", Type = "float", Default = 0.15, Flags = { "SAVE_GAME" } },
		{ Name = "FrontLegFactor", Type = "float", Default = 0.20, Flags = { "SAVE_GAME" } },
		{ Name = "BackLegFactor", Type = "float", Default = 0.05, Flags = { "SAVE_GAME" } },
		{ Name = "FlyingHeight", Type = "float", Default = 40, Flags = { "SAVE_GAME" } },
		{ Name = "ClimbFactor", Type = "float", Default = 0.50, Flags = { "SAVE_GAME" } },
		{ Name = "TurnFactor", Type = "float", Default = 0.01, Flags = { "SAVE_GAME" } },
		{ Name = "Destination", Type = "vec3f", Default = {512, 0, 512}, Flags = { "SAVE_GAME" } },
		{ Name = "DestinationType", Type = "integer", Default = 1, Flags = { "SAVE_GAME" } },
		{ Name = "ReconCount", Type = "integer", Default = 0, Flags = { "SAVE_GAME" } },
		{ Name = "AttackCount", Type = "integer", Default = 0, Flags = { "SAVE_GAME" } },
		{ Name = "FireCount", Type = "integer", Default = 0, Flags = { "SAVE_GAME" } },
    }
}



-- onEnabled ------------------------------------------------------------------------------------------
function COMP_DRAGON:onEnabled()

    self.objDragon = self:getOwner()

    -- Find and record child objects for animation
    self.objDragon:forEachChild(function(child)
        if child.Name == "Wing.Left" then
            self.WingLeft = child
        elseif child.Name == "Wing.Right" then
            self.WingRight = child
        elseif child.Name == "Body" then
            self.Body = child
        elseif child.Name == "Head" then
            self.Head = child
        elseif child.Name == "Tail" then
            self.Tail = child
        elseif child.Name == "Leg.Front.Left" then
            self.LegFrontLeft = child
        elseif child.Name == "Leg.Front.Right" then
            self.LegFrontRight = child
        elseif child.Name == "Leg.Back.Left" then
            self.LegBackLeft = child
        elseif child.Name == "Leg.Back.Right" then
            self.LegBackRight = child
        elseif child.Name == "Emitter" then
            self.Emitter = child
            self.compEmitter = child:getOrCreateComponent("COMP_PARTICLE_EMITTER")
        elseif child.Name == "Target" then
            self.Target = child
        end
    end)

    -- Recalibrate animation
    self:recalibrate()

    -- Set initial count for Recon and Attack
    if not self.InitFlag then

        self.ReconCount = math.random(4, 6)
        self.AttackCount = math.random(3, 5)

        self.InitFlag = true

    end

    mod:log("Dragon has " .. self.ReconCount .. " recons, " .. self.AttackCount .. " attacks.")


end



-- Recalibrate ------------------------------------------------------------------------------------------
function COMP_DRAGON:recalibrate()

    -- Reset tilt on the dragon as a whole
    local posDragon = self.objDragon:getGlobalPosition()
    posDragon[2] = 0
    self.objDragon:globalLookAt(posDragon, true)

    -- Reset tilt on each body part
    self.objDragon:forEachChild(function(child)

        -- reset rotations
        child:setRotationY(0)
        child:setRotationX(0)
        child:setRotationZ(0)

        -- for fire emitter, set fixed angle
        if child.Name == "Emitter" then
            child:setRotationX(145 * math.pi/180)
        end

    end)

    -- Set other variables
    self.timer = -1
    self.fireMode = false
    self.animState = 0
    self.currentAngleWing = 0
    self.currentAngleWhole = 0

end



-- Update ------------------------------------------------------------------------------------------
function COMP_DRAGON:update()

    -- get time elapsed since last frame update
    local dt = self:getLevel():getDeltaTime()
    -- local dt = self:getLevel():getUnscaledDeltaTime() * self:getLevel():getTimeScale()

    -- cap delta time (especially on reload) otherwise animations will go skewed
    if dt > 0 and dt < 0.30 then

        -- Decrement timer
        self.timer = self.timer - dt

        -- Timer: every n seconds, for less frequent calculations
        if self.timer < 0 then

            -- Get distance to destination
            local posDragon = self:getOwner():getGlobalPosition()
            local posDestination = self.Destination
            local dist = math.sqrt( (posDragon[1] - posDestination[1])^2 + (posDragon[3] - posDestination[3])^2 )

            -- Dragon is close to its destination
            if dist < 35 then

                mod:log("Dragon near destination.")
                mod:log("Dragon has " .. self.ReconCount .. " recons, " .. self.AttackCount .. " attacks.")


                -- Collect nearest tribute
                if self.DestinationType == 2 then

                    local objNearestTribute = self:getTribute(posDragon)

                    if objNearestTribute then
                        objNearestTribute:destroy()
                        self.AttackCount = self.AttackCount - 1
                    end

                end


                -- Fly randomly to scare villagers
                if self.ReconCount > 0 then
                    mod:log("Dragon flying to scout the area.")
                    self.Destination = self:getEdgeDestination()
                    self.DestinationType = 1
                    self.timer = 2
                    self.ReconCount = self.ReconCount - 1


                -- Target workplaces to burn
                elseif self.AttackCount > 0 then

                    -- Check for tribute
                    local countTribute = 0
                    self:getLevel():getComponentManager("COMP_DRAGON_TRIBUTE"):getAllEnabledComponent():forEach(function(compTribute)
                        countTribute = countTribute + 1
                    end)

                    -- Fly to tribute
                    if countTribute > 0 then
                        mod:log("Dragon flying to collect tribute.")
                        local objTribute = self:getTribute(posDragon)
                        if objTribute then
                            self.Destination = objTribute:getGlobalPosition()
                            self.DestinationType = 2
                            self.timer = 2
                        end

                    -- Dragon has been blowing fire
                    elseif self.fireMode then
                        mod:log("Dragon finished burning an area, scouting a new area.")
                        self.fireMode = false
                        self.timer = 2
                        if self.DestinationType == 1 then
                            self.Destination = self:getWorkplaceDestination()
                            self.DestinationType = 3
                        else
                            self.Destination = self:getEdgeDestination()
                            self.DestinationType = 1
                        end
                        self.AttackCount = self.AttackCount - 1

                    -- Dragon arrived at a workplace and should start blowing fire
                    elseif self.DestinationType == 3 then
                        mod:log("Dragon started burning an area.")
                        self.fireMode = true
                        self.timer = 5

                        local posBurn = self:raycastFire()
                        if posBurn then
                            local objBurn = self:getLevel():createObject("PREFAB_FIRE_START", posBurn)
                            objBurn:setScale(0)
                            local compFire = objBurn:getOrCreateComponent("COMP_FIRE")
                            compFire.TimerList[1] = 22 -- shorten the first fire object's timer so it spreads quickly, 20 seconds is the spread timer
                            compFire.EachSpreadDistance = 7 -- make the spread wider than normal
                            self.FireCount = self.FireCount + 1
                        end

                        -- Notification that a fire has started
                        local compVillager = self:getLevel():getComponentManager("COMP_VILLAGER"):getFirstEnabled()
                        local jobAsset = nil
                        local jobWorkplace = nil
                        if (compVillager) then

                            local job = compVillager:getJobInstance()
                            if (job) then
                                jobAsset = job.AssetJob
                                jobWorkplace = job.Workplace
                            end

                            compVillager:changeJob("JOB_FIRE_EVENT_START")

                            if (job) then
                                if (jobWorkplace) then
                                    compVillager:changeJob(jobAsset, jobWorkplace)
                                else
                                    compVillager:changeJob(jobAsset)
                                end
                            else
                                compVillager:changeJob(nil)
                            end

                        end -- Villager found to trigger the notification

                    else
                        mod:log("Dragon targeting next workplace.")
                        self.Destination = self:getWorkplaceDestination()
                        self.DestinationType = 3
                        self.timer = 2

                    end


                -- Leave the map
                elseif self.AttackCount == 0 then
                    -- mod:log("Dragon is departing.")
                    self.fireMode = false
                    self.Destination = {0, 0, 0}
                    self.DestinationType = 0
                    self.timer = 2
                    self.AttackCount = self.AttackCount - 1


                -- Disappear
                else

                    -- Get reward if no fires were started
                    if self.FireCount == 0 then

                        local compVillager = self:getLevel():getComponentManager("COMP_VILLAGER"):getFirstEnabled()
                        local jobAsset = nil
                        local jobWorkplace = nil
                        if (compVillager) then

                            local job = compVillager:getJobInstance()
                            if (job) then
                                jobAsset = job.AssetJob
                                jobWorkplace = job.Workplace
                            end

                            compVillager:changeJob("JOB_DRAGON_EVENT_SUCCESS")

                            if (job) then
                                if (jobWorkplace) then
                                    compVillager:changeJob(jobAsset, jobWorkplace)
                                else
                                    compVillager:changeJob(jobAsset)
                                end
                            else
                                compVillager:changeJob(nil)
                            end

                        end -- Villager found to trigger the notification

                    end


                    -- Destroy the dragon, which will destroy this component and stop all code
                    self:getOwner():destroy()

                end


            elseif self.fireMode then
                self.fireMode = false
                self.timer = 2

            else
                self.timer = 2

            end


        end

        self:flyDragon(dt)

    end

end


-- Raycast Height
function COMP_DRAGON:raycastHeight(pos)

    local raycastResult = {}
    local FromPosition = { pos[1], pos[2]+100, pos[3] }
    local ToPosition = { pos[1], pos[2]-100, pos[3] }

    local flag = bit.bor(
        bit.lshift(1, OBJECT_FLAG.TERRAIN:toNumber()),
        bit.lshift(1, OBJECT_FLAG.WATER:toNumber()),
        bit.lshift(1, OBJECT_FLAG.PLATFORM:toNumber()),
        bit.lshift(1, OBJECT_FLAG.BRIDGE:toNumber()),
        bit.lshift(1, OBJECT_FLAG.BUILDING_PART:toNumber()),
        bit.lshift(1, OBJECT_FLAG.BUILDING_PART_NON_INTERACTIVE:toNumber()),
        bit.lshift(1, OBJECT_FLAG.WALL_PART:toNumber())
    )

    self:getLevel():rayCast(FromPosition, ToPosition, raycastResult, flag)

    return raycastResult.Position.y

end


-- Raycast Fire
function COMP_DRAGON:raycastFire()

    local posEmitter = self.Emitter:getGlobalPosition()
    local posTarget = self.Target:getGlobalPosition()

    local raycastResult = {}
    local FromPosition = { posEmitter[1], posEmitter[2], posEmitter[3] }
    local ToPosition = { posTarget[1], posTarget[2], posTarget[3] }

    local flag = bit.bor(
        bit.lshift(1, OBJECT_FLAG.TERRAIN:toNumber()),
        bit.lshift(1, OBJECT_FLAG.WATER:toNumber()),
        bit.lshift(1, OBJECT_FLAG.PLATFORM:toNumber()),
        bit.lshift(1, OBJECT_FLAG.BRIDGE:toNumber()),
        bit.lshift(1, OBJECT_FLAG.BUILDING_PART:toNumber()),
        bit.lshift(1, OBJECT_FLAG.BUILDING_PART_NON_INTERACTIVE:toNumber()),
        bit.lshift(1, OBJECT_FLAG.WALL_PART:toNumber())
    )

    self:getLevel():rayCast(FromPosition, ToPosition, raycastResult, flag)

    return raycastResult.Position

end


-- Fly dragon
function COMP_DRAGON:flyDragon(dt)

    if dt == 0 then -- game is paused

    else

        -- current position
        local posDragon = self.objDragon:getGlobalPosition()
        local state = self.animState
        self.ground = posDragon[2] - self:raycastHeight(posDragon)

        -- current destination
        local posDestination = self.Destination

        -- ease the rotations at the top and bottom of the movement
        local ease = easeValue(self.currentAngleWing)

        -- calculate movement speed amount based on time elapsed
        local moveForward = dt * self.MovementSpeed

        -- calculate body part rotation amount based on time elapsed
        local rotation = dt * self.AnimationSpeed * ease

        -- mod:log("dt " .. dt .. ", AnimationSpeed " .. self.AnimationSpeed .. ", ease " .. ease .. ", rotation " .. rotation .. ", state " .. state)
        -- mod:log("Destination " .. tostring(self.Destination))
        -- mod:log("state " .. state .. ", ease .. " .. ease .. ", currentAngleWing " .. self.currentAngleWing .. ", max in state 1 " .. (self.WingAngle * math.pi/180) .. ", min in state -1 " .. (-self.WingAngle * math.pi/180))

        if (state == 0) then

            self:recalibrate()
            state = 1


        elseif (state == 1) and (self.currentAngleWing < (self.WingAngle * math.pi/180)) then -- move body parts one direction

            self:animate(rotation, state)


        elseif (state == 1) then -- reached the end of the stroke, reverse directions

            state = -1


        elseif (state == -1) and (self.currentAngleWing > (-self.WingAngle * math.pi/180)) then -- move body parts another direction

            self:animate(rotation, state)


        elseif (state == -1) then -- reached the end of the stroke, reverse directions

            state = 1


        end


        -- Calculate change in altitude required to meet ideal flying height
        local moveVertical = dt * (self.FlyingHeight - self.ground) * self.ClimbFactor


        -- Breath fire and move
        if self.fireMode then -- Dragon is breathing fire

            if self.compEmitter and self.timer > 3.0 then
                self.compEmitter.IsEmitting = true
            else -- turn off fire before you move the dragon, to avoid the fire following
                self.compEmitter.IsEmitting = false
            end

        else -- Dragon is not breathing fire and should move forward and rotate

            if self.compEmitter then
                self.compEmitter.IsEmitting = false
            end

            -- Move forward
            self.objDragon:move({0, moveVertical, moveForward})

            -- Rotate toward the destination
            -- mod:log("posDragon3 " .. posDragon[3] .. ", destination3 " .. self.Destination[3])
            -- mod:log("posDragon1 " .. posDragon[1] .. ", destination1 " .. self.Destination[1])

            local diffY = posDragon[3] - self.Destination[3]
            local diffX = posDragon[1] - self.Destination[1]
            -- mod:log("diffY " .. diffY .. ", diffX " .. diffX)

            -- arc tangent (angle from x axis to the destination)
            local angleDestination = (math.atan2(diffX, diffY) * 180/math.pi)

            -- find the dragon's current vector
            local angleDragon = math.deg(quaternion.getEulerAngles(self.objDragon:getGlobalOrientation())[2])
            angleDragon = angleDragon + 180
            angleDragon = math.fmod(angleDragon, 360)

            -- difference in angles
            local angleDifference = math.fmod(angleDestination - angleDragon + 900, 360) - 180

            -- add factors to rotation
            angleDifference = angleDifference * dt * self.TurnFactor

            -- rotate the dragon
            self.objDragon:rotateLocalY(angleDifference)

        end

        -- record final state
        self.animState = state

        -- mod:log("angleDestination before " .. (math.atan2(diffY, diffX) * 180/math.pi) .. ", angleDestination after " .. angleDestination .. ", angleDragon before " .. math.deg(quaternion.getEulerAngles(self.objDragon:getGlobalOrientation())[2]) .. ", angleDragon after " .. angleDragon .. ", rotation " .. rotation)

    end

end


function COMP_DRAGON:animate(rotation, state)

    self.WingLeft:rotateZ(state * rotation)
    self.WingRight:rotateZ(state * -rotation)
    self.currentAngleWing = self.currentAngleWing + (state * rotation)

    self.objDragon:rotateLocalX(state * rotation * self.WholeFactor)
    self.currentAngleWhole = self.currentAngleWhole + (state * rotation * self.WholeFactor)

    self.Emitter:rotateX(state * -rotation * self.WholeFactor)

    self.Head:rotateX(state * -rotation * self.HeadFactor)

    self.Tail:rotateZ(state * rotation * self.TailFactor)

    self.LegFrontRight:rotateZ(state * rotation * self.FrontLegFactor)
    self.LegFrontLeft:rotateX(state * rotation * self.FrontLegFactor)
    self.LegBackRight:rotateX(state * rotation * self.BackLegFactor)
    self.LegBackLeft:rotateZ(state * rotation * self.BackLegFactor)

end


-- Get Random edge destination
function COMP_DRAGON:getEdgeDestination()

    local posX = math.random(124, 900)
    local posY = math.random(124, 900)

    return {posX, 0, posY}

end


-- Get Workplace Destination
function COMP_DRAGON:getWorkplaceDestination()

    local posWork = nil
    local compWorkClosest = nil
    local dist = nil

    local posDragon = self:getOwner():getGlobalPosition()
    local posTarget = nil
    local compWorkName = nil
    local listWork = nil
    listWork = {}

    -- Find all workplaces and record with distance
    self:getLevel():getComponentManager("COMP_WORKPLACE"):getAllEnabledComponent():forEach(function(compWork)

        -- do not record basic workplaces that tend to be far away from town, and have nothing interesting to destroy
        compWorkName = compWork:getOwner().Name
        if (compWorkName == "ForesterPart" or compWorkName == "WoodCutterCampPart" or compWorkName == "StonecutterPart" or compWorkName == "HunterPart") then
        else

            posWork = compWork:getOwner():getGlobalPosition()
            dist = math.sqrt( (posDragon.x - posWork.x)^2 + (posDragon.z - posWork.z)^2 )

            table.insert(listWork, {compWork = compWork, dist = dist})

        end

    end)

    -- Sort the workplaces by distance
    table.sort(listWork, SortByDist)

    -- Take a random one in the closest five workplaces
    local topN = 5
    if #listWork > 0 then

        if (#listWork < 5) then
            topN = #listWork
        end

        local rand = math.random(1, topN)
        compWorkClosest = listWork[rand].compWork

        -- Get the location of this workplace
        if (compWorkClosest) then

            posTarget = compWorkClosest:getOwner():getGlobalPosition()

            if (posTarget) then
                return posTarget
            else
                return nil
            end

        end

    -- There are no worklaces
    else

        local edgeDest = self:getEdgeDestination()
        return edgeDest

    end

end


-- Get Tribute Destination
function COMP_DRAGON:getTribute(posDragon)

    local distNearestTribute = 1000
    local objNearestTribute = nil
    self:getLevel():getComponentManager("COMP_DRAGON_TRIBUTE"):getAllEnabledComponent():forEach(function(compTribute)

        local objTribute = compTribute:getOwner()
        local posTribute = objTribute:getGlobalPosition()
        local dist = math.sqrt( (posDragon[1] - posTribute[1])^2 + (posDragon[3] - posTribute[3])^2 )

        if dist < distNearestTribute then
            distNearestTribute = dist
            objNearestTribute = objTribute
        end

    end)

    return objNearestTribute

end



-- Register
mod:registerClass(COMP_DRAGON)

-- Attach component
-- mod:registerPrefabComponent("PREFAB_DRAGON_DRAGON1", { DataType = "COMP_AGENT", BehaviorTree = "BEHAVIOR_DRAGON", RoadPaintSpeed = 0.0 } )
mod:registerPrefabComponent("PREFAB_DRAGON_DRAGON1", { DataType = "COMP_DRAGON" } )

